from sklearn.linear_model import *
from finol.utils import load_config

distiller_dict = {
    ###############################
    # Classical linear regressors #
    ###############################
    "LinearRegression": LinearRegression,  # Ordinary least squares Linear Regression
    "Ridge": Ridge,  # Linear least squares with l2 regularization
    "RidgeCV": RidgeCV,  # Ridge regression with built-in cross-validation
    "SGDRegressor": SGDRegressor,  # Linear model fitted by minimizing a regularized empirical loss with SGD

    ######################################
    # Regressors with variable selection #
    ######################################
    "ElasticNet": ElasticNet,  # Linear regression with combined L1 and L2 priors as regularizer
    "ElasticNetCV": ElasticNetCV,  # Elastic Net model with iterative fitting along a regularization path
    "Lars": Lars,  # Least Angle Regression model a.k.a
    "LarsCV": LarsCV,  # Cross-validated Least Angle Regression model
    "Lasso": Lasso,  # Linear Model trained with L1 prior as regularizer (aka the Lasso)
    "LassoCV": LassoCV,  # Lasso linear model with iterative fitting along a regularization path
    "LassoLars": LassoLars,  # Lasso model fit with Least Angle Regression a.k.a
    "LassoLarsCV": LassoLarsCV,  # Cross-validated Lasso, using the LARS algorithm
    "LassoLarsIC": LassoLarsIC,  # Lasso model fit with Lars using BIC or AIC for model selection
    "OrthogonalMatchingPursuit": OrthogonalMatchingPursuit,  # Orthogonal Matching Pursuit model (OMP)
    "OrthogonalMatchingPursuitCV": OrthogonalMatchingPursuitCV,  # Cross-validated Orthogonal Matching Pursuit model (OMP)

    #######################
    # Bayesian regressors #
    #######################
    "ARDRegression": ARDRegression,  # Bayesian ARD regression
    "BayesianRidge": BayesianRidge,  # Bayesian ridge regression

    #############################
    # Outlier-robust regressors #
    #############################
    "HuberRegressor": HuberRegressor,  # L2-regularized linear regression model that is robust to outliers
    "QuantileRegressor": QuantileRegressor,  # Linear regression model that predicts conditional quantiles
    "RANSACRegressor": RANSACRegressor,  # RANSAC (RANdom SAmple Consensus) algorithm
    "TheilSenRegressor": TheilSenRegressor,  # Theil-Sen Estimator: robust multivariate regression model

    ##################################################
    # Generalized linear models (GLM) for regression #
    ##################################################
    "PoissonRegressor": PoissonRegressor,  # Generalized Linear Model with a Poisson distribution
    "TweedieRegressor": TweedieRegressor,  # Generalized Linear Model with a Tweedie distribution
    "GammaRegressor": GammaRegressor,  # Generalized Linear Model with a Gamma distribution

    #################
    # Miscellaneous #
    #################
    "PassiveAggressiveRegressor": PassiveAggressiveRegressor,  # Passive Aggressive Regressor
}


class DistillerSelector:
    def __init__(self):
        self.config = load_config()

    def select_distiller(self):
        distiller_cls = distiller_dict.get(self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["DISTILLER_NAME"], None)
        if distiller_cls is None:
            raise ValueError(f"Invalid distiller: {self.config['INTERPRETABLE_ANALYSIS_CONFIG']['DISTILLER_NAME']}. Supported distillers are: {distiller_dict}")
        return distiller_cls()




import time
import json
import tkinter as tk
from tkinter import messagebox, ttk
from finol.data_layer.DatasetLoader import DatasetLoader
from finol.optimization_layer.ModelTrainer import ModelTrainer
from finol.evaluation_layer.ModelEvaluator import ModelEvaluator
from finol.utils import *


# def format_string(input_string):
#     words = input_string.split("_")
#     formatted_words = [word.capitalize() for word in words]
#     formatted_string = " ".join(formatted_words)
#
#     return formatted_string


class FinOLApp:
    def __init__(self, root):
        self.root = root
        self.root.title("FinOL: Towards Open Benchmarking for Data-Driven Online Portfolio Selection")
        self.create_widgets()

        self.experiment_details = {}

    def create_widgets(self):
        ###############################
        # General Layer Configuration #
        ###############################
        self.general_config_frame = tk.LabelFrame(self.root, text="General Configuration", font=("Helvetica", 10, "bold"))
        self.general_config_frame.pack(padx=10, pady=1, fill="none")

        # GET_LATEST_FINOL
        self.GET_LATEST_FINOL_var = tk.BooleanVar(value=False)
        GET_LATEST_FINOL_chk = tk.Checkbutton(self.general_config_frame, text="Get Latest FinOL", variable=self.GET_LATEST_FINOL_var)
        GET_LATEST_FINOL_chk.grid(row=0, column=0, padx=10, pady=1)

        # DEVICE
        self.DEVICE_options = ["cpu", "cuda"]
        tk.Label(self.general_config_frame, text="Select Device:").grid(row=0, column=1, padx=10, pady=1)
        self.DEVICE_var = tk.StringVar()
        self.DEVICE_dropdown = ttk.Combobox(self.general_config_frame, textvariable=self.DEVICE_var, values=self.DEVICE_options)
        self.DEVICE_dropdown.grid(row=0, column=2, padx=10, pady=1)
        self.DEVICE_dropdown.current(0)

        # MANUAL_SEED
        tk.Label(self.general_config_frame, text="Set Seed:").grid(row=0, column=3, padx=10, pady=1)
        self.MANUAL_SEED_var = tk.IntVar()
        self.MANUAL_SEED_entry = tk.Entry(self.general_config_frame, textvariable=self.MANUAL_SEED_var)
        self.MANUAL_SEED_entry.grid(row=0, column=4, padx=10, pady=1)
        self.MANUAL_SEED_var.set(0)


        ############################
        # Data Layer Configuration #
        ############################
        self.data_config_frame = tk.LabelFrame(self.root, text="Data Layer Configuration", font=("Helvetica", 10, "bold"))
        self.data_config_frame.pack(padx=10, pady=1, fill="none")

        # DATASET_NAME
        self.DATASET_NAME_options = ["NYSE(O)", "NYSE(N)", "DJIA", "SP500", "TSE", "SSE", "HSI", "CMEG", "CRYPTO"]
        tk.Label(self.data_config_frame, text="Select Dataset:").grid(row=0, column=0, padx=10, pady=1)
        self.DATASET_NAME_var = tk.StringVar()
        self.DATASET_NAME_dropdown = ttk.Combobox(self.data_config_frame, textvariable=self.DATASET_NAME_var, values=self.DATASET_NAME_options)
        self.DATASET_NAME_dropdown.grid(row=0, column=1, padx=10, pady=1)
        self.DATASET_NAME_dropdown.current(0)

        # SCALER
        self.SCALER_options = ["None", "StandardScaler", "MinMaxScaler", "MaxAbsScaler", "RobustScaler",
                       "WindowStandardScaler", "WindowMinMaxScaler", "WindowMaxAbsScaler", "WindowRobustScaler"]
        tk.Label(self.data_config_frame, text="Select Scaler:").grid(row=0, column=2, padx=10, pady=1)
        self.SCALER_var = tk.StringVar()
        self.SCALER_dropdown = ttk.Combobox(self.data_config_frame, textvariable=self.SCALER_var, values=self.SCALER_options)
        self.SCALER_dropdown.grid(row=0, column=3, padx=10, pady=1)
        self.SCALER_dropdown.current(5)

        separator = ttk.Separator(self.data_config_frame,  orient="horizontal")
        separator.grid(row=1, column=0, columnspan=4, padx=10, pady=1, sticky="ew")
        separator_label = ttk.Label(self.data_config_frame, text="Auto Feature Engineering")
        separator_label.grid(row=1, column=0, columnspan=4, padx=10, pady=1)

        # INCLUDE_OHLCV_FEATURES
        self.INCLUDE_OHLCV_FEATURES_var = tk.BooleanVar(value=True)
        self.INCLUDE_OHLCV_FEATURES_chk = tk.Checkbutton(self.data_config_frame, text="Include OHLCV Features", variable=self.INCLUDE_OHLCV_FEATURES_var)
        self.INCLUDE_OHLCV_FEATURES_chk.grid(row=2, column=0)

        # INCLUDE_OVERLAP_FEATURES
        self.INCLUDE_OVERLAP_FEATURES_var = tk.BooleanVar(value=True)
        self.INCLUDE_OVERLAP_FEATURES_chk = tk.Checkbutton(self.data_config_frame, text="Include Overlap Features", variable=self.INCLUDE_OVERLAP_FEATURES_var)
        self.INCLUDE_OVERLAP_FEATURES_chk.grid(row=2, column=1)

        # INCLUDE_MOMENTUM_FEATURES
        self.INCLUDE_MOMENTUM_FEATURES_var = tk.BooleanVar(value=True)
        self.INCLUDE_MOMENTUM_FEATURES_chk = tk.Checkbutton(self.data_config_frame, text="Include Momentum Features", variable=self.INCLUDE_MOMENTUM_FEATURES_var)
        self.INCLUDE_MOMENTUM_FEATURES_chk.grid(row=2, column=2)

        # INCLUDE_VOLUME_FEATURES
        self.INCLUDE_VOLUME_FEATURES_var = tk.BooleanVar(value=True)
        self.INCLUDE_VOLUME_FEATURES_chk = tk.Checkbutton(self.data_config_frame, text="Include Volume Features", variable=self.INCLUDE_VOLUME_FEATURES_var)
        self.INCLUDE_VOLUME_FEATURES_chk.grid(row=2, column=3)

        # INCLUDE_CYCLE_FEATURES
        self.INCLUDE_CYCLE_FEATURES_var = tk.BooleanVar(value=True)
        self.INCLUDE_CYCLE_FEATURES_chk = tk.Checkbutton(self.data_config_frame, text="Include Cycle Features", variable=self.INCLUDE_CYCLE_FEATURES_var)
        self.INCLUDE_CYCLE_FEATURES_chk.grid(row=3, column=0)

        #  INCLUDE_PRICE_FEATURES
        self.INCLUDE_PRICE_FEATURES_var = tk.BooleanVar(value=True)
        self.INCLUDE_PRICE_FEATURES_chk = tk.Checkbutton(self.data_config_frame, text="Include Price Features", variable=self.INCLUDE_PRICE_FEATURES_var)
        self.INCLUDE_PRICE_FEATURES_chk.grid(row=3, column=1)

        # INCLUDE_VOLATILITY_FEATURES
        self.INCLUDE_VOLATILITY_FEATURES_var = tk.BooleanVar(value=True)
        self.INCLUDE_VOLATILITY_FEATURES_chk = tk.Checkbutton(self.data_config_frame, text="Include Volatility Features", variable=self.INCLUDE_VOLATILITY_FEATURES_var)
        self.INCLUDE_VOLATILITY_FEATURES_chk.grid(row=3, column=2)

        #  INCLUDE_PATTERN_FEATURES
        self.INCLUDE_PATTERN_FEATURES_var = tk.BooleanVar(value=True)
        self.INCLUDE_PATTERN_FEATURES_chk = tk.Checkbutton(self.data_config_frame, text="Include Pattern Features", variable=self.INCLUDE_PATTERN_FEATURES_var)
        self.INCLUDE_PATTERN_FEATURES_chk.grid(row=3, column=3)

        separator = ttk.Separator(self.data_config_frame,  orient="horizontal")
        separator.grid(row=4, column=0, columnspan=4, padx=10, pady=1, sticky="ew")
        separator_label = ttk.Label(self.data_config_frame, text="Data Augmentation")
        separator_label.grid(row=4, column=0, columnspan=4, padx=10, pady=1)

        # INCLUDE_WINDOW_DATA
        self.INCLUDE_WINDOW_DATA_var = tk.BooleanVar(value=True)
        self.INCLUDE_WINDOW_DATA_chk = tk.Checkbutton(self.data_config_frame, text="Include Window Data", variable=self.INCLUDE_WINDOW_DATA_var, command=self.base_on_INCLUDE_WINDOW_DATA)
        self.INCLUDE_WINDOW_DATA_chk.grid(row=5, column=0)

        # WINDOW_SIZE
        tk.Label(self.data_config_frame, text="Set Window Size:").grid(row=5, column=1, padx=10, pady=1)
        self.WINDOW_SIZE_var = tk.IntVar()
        self.WINDOW_SIZE_entry = tk.Entry(self.data_config_frame, textvariable=self.WINDOW_SIZE_var)
        self.WINDOW_SIZE_entry.grid(row=5, column=2)
        self.WINDOW_SIZE_var.set(10)

        separator = ttk.Separator(self.data_config_frame,  orient="horizontal")
        separator.grid(row=6, column=0, columnspan=4, padx=10, pady=1, sticky="ew")

        # LOAD_LOCAL_DATALOADER
        self.LOAD_LOCAL_DATALOADER_var = tk.BooleanVar(value=False)
        self.LOAD_LOCAL_DATALOADER_checkbox = tk.Checkbutton(self.data_config_frame, text="Load Local Dataloader", variable=self.LOAD_LOCAL_DATALOADER_var, command=self.base_on_LOAD_LOCAL_DATALOADER)
        self.LOAD_LOCAL_DATALOADER_checkbox.grid(row=7, column=0, padx=10, pady=1)

        # load_button
        self.load_button = tk.Button(self.root, text="Load Dataset", command=self.load_dataset)
        self.load_button.pack(padx=10, pady=3)

        #############################
        # Model Layer Configuration #
        #############################
        self.model_config_frame = tk.LabelFrame(self.root, text="Model Layer Configuration", font=("Helvetica", 10, "bold"))
        self.model_config_frame.pack(padx=10, pady=1, fill="none")

        # ttk.Label(self.model_config_frame, text=" "*150).grid(row=0, column=0, columnspan=4, padx=10, pady=0)
        # ttk.Label(self.model_config_frame, text=" "*150).grid(row=100, column=0, columnspan=4, padx=10, pady=0)

        # MODEL_NAME
        self.MODEL_NAME_options = ["DNN", "RNN", "LSTM", "CNN", "Transformer", "LSRE-CAAN", "LSRE-CAAN-d", "LSRE-CAAN-dd",
                                   "AlphaPortfolio"]
        tk.Label(self.model_config_frame, text="Select Model:").grid(row=1, column=0, padx=10, pady=1)
        self.MODEL_NAME_var = tk.StringVar()
        self.MODEL_NAME_dropdown = ttk.Combobox(self.model_config_frame, textvariable=self.MODEL_NAME_var, values=self.MODEL_NAME_options)
        self.MODEL_NAME_dropdown.grid(row=1, column=1, padx=10, pady=1)
        self.MODEL_NAME_dropdown.current(0)

        separator = ttk.Separator(self.model_config_frame,  orient="horizontal")
        separator.grid(row=2, column=0, columnspan=4, padx=10, pady=1, sticky="ew")
        separator_label = ttk.Label(self.model_config_frame, text="Auto Hyper-parameters Tuning")
        separator_label.grid(row=2, column=0, columnspan=4, padx=10, pady=1)

        # TUNE_PARAMETERS
        self.TUNE_PARAMETERS_var = tk.BooleanVar(value=False)
        self.TUNE_PARAMETERS_checkbox = tk.Checkbutton(self.model_config_frame, text="Tune Hyper-parameters", variable=self.TUNE_PARAMETERS_var, command=self.base_on_TUNE_PARAMETERS)
        self.TUNE_PARAMETERS_checkbox.grid(row=3, column=0, padx=10, pady=1)

        # NUM_TRIALS
        tk.Label(self.model_config_frame, text="Number of Trials:").grid(row=3, column=1, padx=10, pady=1)
        self.NUM_TRIALS_var = tk.IntVar()
        self.NUM_TRIALS_entry = tk.Entry(self.model_config_frame, textvariable=self.NUM_TRIALS_var)
        self.NUM_TRIALS_entry.grid(row=3, column=2)
        self.NUM_TRIALS_var.set(20)

        ####################################
        # Optimization Layer Configuration #
        ####################################
        self.optimization_config_frame = tk.LabelFrame(self.root, text="Optimization Layer Configuration", font=("Helvetica", 10, "bold"))
        self.optimization_config_frame.pack(padx=10, pady=1, fill="none")

        # ttk.Label(self.optimization_config_frame, text=" "*150).grid(row=0, column=0, columnspan=4, padx=10, pady=0)
        # ttk.Label(self.optimization_config_frame, text=" "*150).grid(row=1000, column=0, columnspan=4, padx=10, pady=0)

        # NUM_EPOCHES
        tk.Label(self.optimization_config_frame, text="Number of Epoches:").grid(row=1, column=0, padx=10, pady=1)
        self.NUM_EPOCHES_var = tk.IntVar()
        self.NUM_EPOCHES_entry = tk.Entry(self.optimization_config_frame, textvariable=self.NUM_EPOCHES_var)
        self.NUM_EPOCHES_entry.grid(row=1, column=1, padx=10, pady=1)
        self.NUM_EPOCHES_var.set(200)

        # SAVE_EVERY
        tk.Label(self.optimization_config_frame, text="Save Every:").grid(row=1, column=2, padx=10, pady=1)
        self.SAVE_EVERY_var = tk.IntVar()
        self.SAVE_EVERY_entry = tk.Entry(self.optimization_config_frame, textvariable=self.SAVE_EVERY_var)
        self.SAVE_EVERY_entry.grid(row=1, column=3, padx=10, pady=1)
        self.SAVE_EVERY_var.set(1)

        separator = ttk.Separator(self.optimization_config_frame,  orient="horizontal")
        separator.grid(row=2, column=0, columnspan=4, padx=10, pady=1, sticky="ew")
        separator_label = ttk.Label(self.optimization_config_frame, text="Optimizer Settings")
        separator_label.grid(row=2, column=0, columnspan=4, padx=10, pady=1)

        # OPTIMIZER_NAME
        self.OPTIMIZER_NAME_options = ["Adadelta", "Adagrad", "Adam", "AdamW", "Adamax", "ASGD", "SGD", "RAdam", "Rprop",
                                       "RMSprop", "NAdam", "A2GradExp", "A2GradInc", "A2GradUni", "AccSGD", "AdaBelief",
                                       "AdaBound", "AdaMod", "Adafactor", "AdamP", "AggMo", "Apollo", "DiffGrad", "LARS",
                                       "Lamb", "MADGRAD", "NovoGrad", "PID", "QHAdam", "QHM", "Ranger", "RangerQH", "RangerVA",
                                       "SGDP", "SGDW", "SWATS", "Yogi"]
        tk.Label(self.optimization_config_frame, text="Select Optimizer:").grid(row=3, column=0, padx=10, pady=1)
        self.OPTIMIZER_NAME_var = tk.StringVar()
        self.OPTIMIZER_NAME_dropdown = ttk.Combobox(self.optimization_config_frame, textvariable=self.OPTIMIZER_NAME_var, values=self.OPTIMIZER_NAME_options)
        self.OPTIMIZER_NAME_dropdown.grid(row=3, column=1, padx=10, pady=1)
        self.OPTIMIZER_NAME_dropdown.current(2)

        # LEARNING_RATE
        tk.Label(self.optimization_config_frame, text="Set Learning Rate:").grid(row=3, column=2, padx=10, pady=1)
        self.LEARNING_RATE_var = tk.DoubleVar()
        self.LEARNING_RATE_entry = tk.Entry(self.optimization_config_frame, textvariable=self.LEARNING_RATE_var)
        self.LEARNING_RATE_entry.grid(row=3, column=3, padx=10, pady=1)
        self.LEARNING_RATE_var.set(0.001)

        separator = ttk.Separator(self.optimization_config_frame,  orient="horizontal")
        separator.grid(row=4, column=0, columnspan=4, padx=10, pady=1, sticky="ew")
        separator_label = ttk.Label(self.optimization_config_frame, text="Criterion Settings")
        separator_label.grid(row=4, column=0, columnspan=4, padx=10, pady=1)

        # CRITERION_NAME
        self.CRITERION_NAME_options = ["LOG_WEALTH", "LOG_WEALTH_L2_DIVERSIFICATION", "LOG_WEALTH_L2_CONCENTRATION",
                                       "L2_DIVERSIFICATION", "L2_CONCENTRATION", "SHARPE_RATIO", "VOLATILITY"]
        tk.Label(self.optimization_config_frame, text="Select Criterion:").grid(row=6, column=0, padx=10, pady=1)
        self.CRITERION_NAME_var = tk.StringVar()
        self.CRITERION_NAME_dropdown = ttk.Combobox(self.optimization_config_frame, textvariable=self.CRITERION_NAME_var, values=self.CRITERION_NAME_options)
        self.CRITERION_NAME_dropdown.grid(row=6, column=1, padx=10, pady=1)
        self.CRITERION_NAME_dropdown.current(0)

        # LAMBDA_L2
        tk.Label(self.optimization_config_frame, text="Set Lambda:").grid(row=6, column=2, padx=10, pady=1)
        self.LAMBDA_L2_var = tk.DoubleVar()
        self.LAMBDA_L2_entry = tk.Entry(self.optimization_config_frame, textvariable=self.LAMBDA_L2_var)
        self.LAMBDA_L2_entry.grid(row=6, column=3, padx=10, pady=1)
        self.LAMBDA_L2_var.set(0.0005)

        # Trace changes to CRITERION_NAME_var
        self.CRITERION_NAME_var.trace("w", self.base_on_CRITERION_NAME)
        # Initial state update
        self.base_on_CRITERION_NAME()

        # train_button
        self.train_button = tk.Button(self.root, text="Train Model", command=self.train_model)
        self.train_button.pack(padx=10, pady=3)

        ##################################
        # Evaluation Layer Configuration #
        ##################################
        self.evaluation_config_frame = tk.LabelFrame(self.root, text="Evaluation Layer Configuration", font=("Helvetica", 10, "bold"))
        self.evaluation_config_frame.pack(padx=10, pady=1, fill="none")

        # PLOT_CHINESE
        self.PLOT_CHINESE_var = tk.BooleanVar(value=False)
        self.PLOT_CHINESE_checkbox = tk.Checkbutton(self.evaluation_config_frame, text="Plot Chinese", variable=self.PLOT_CHINESE_var)
        self.PLOT_CHINESE_checkbox.grid(row=0, column=0, padx=10, pady=1)

        # PROP_WINNERS
        tk.Label(self.evaluation_config_frame, text="Proportion of Winners").grid(row=0, column=1, padx=10, pady=1)
        self.PROP_WINNERS_var = tk.DoubleVar()
        self.PROP_WINNERS_entry = tk.Entry(self.evaluation_config_frame, textvariable=self.PROP_WINNERS_var)
        self.PROP_WINNERS_entry.grid(row=0, column=2, padx=10, pady=1)
        self.PROP_WINNERS_var.set(0.5)

        separator = ttk.Separator(self.evaluation_config_frame,  orient="horizontal")
        separator.grid(row=1, column=0, columnspan=4, padx=10, pady=1, sticky="ew")
        separator_label = ttk.Label(self.evaluation_config_frame, text="Interpretability Analysis")
        separator_label.grid(row=1, column=0, columnspan=4, padx=10, pady=1)

        # INCLUDE_INTERPRETABILITY_ANALYSIS
        self.INCLUDE_INTERPRETABILITY_ANALYSIS_var = tk.BooleanVar(value=False)
        self.INCLUDE_INTERPRETABILITY_ANALYSIS_checkbox = tk.Checkbutton(self.evaluation_config_frame, text="Include Interpretability Analysis", variable=self.INCLUDE_INTERPRETABILITY_ANALYSIS_var)
        self.INCLUDE_INTERPRETABILITY_ANALYSIS_checkbox.grid(row=2, column=0, padx=10, pady=1)

        # INCLUDE_ECONOMIC_DISTILLATION
        self.INCLUDE_ECONOMIC_DISTILLATION_var = tk.BooleanVar(value=False)
        self.INCLUDE_ECONOMIC_DISTILLATION_checkbox = tk.Checkbutton(self.evaluation_config_frame, text="Include Economic Distillation", variable=self.INCLUDE_ECONOMIC_DISTILLATION_var)
        self.INCLUDE_ECONOMIC_DISTILLATION_checkbox.grid(row=2, column=1, padx=10, pady=1)

        # PROP_DISTILLED_FEATURES
        tk.Label(self.evaluation_config_frame, text="Proportion of Distilled Features:").grid(row=2, column=2, padx=10, pady=1)
        self.PROP_DISTILLED_FEATURES_var = tk.DoubleVar()
        self.PROP_DISTILLED_FEATURES_entry = tk.Entry(self.evaluation_config_frame, textvariable=self.PROP_DISTILLED_FEATURES_var)
        self.PROP_DISTILLED_FEATURES_entry.grid(row=2, column=3, padx=10, pady=1)
        self.PROP_DISTILLED_FEATURES_var.set(0.7)

        # DISTILLER_NAME
        self.DISTILLER_NAME_options = ["LinearRegression", "Ridge", "RidgeCV", "SGDRegressor", "ElasticNet", "ElasticNetCV",
                                       "Lars", "LarsCV", "Lasso", "LassoCV", "LassoLars", "LassoLarsCV", "LassoLarsIC",
                                       "OrthogonalMatchingPursuit", "OrthogonalMatchingPursuitCV", "ARDRegression",
                                       "BayesianRidge", "HuberRegressor", "QuantileRegressor", "RANSACRegressor",
                                       "TheilSenRegressor", "PoissonRegressor", "TweedieRegressor", "GammaRegressor",
                                       "PassiveAggressiveRegressor"]
        tk.Label(self.evaluation_config_frame, text="Select Distiller:").grid(row=3, column=0, padx=10, pady=1)
        self.DISTILLER_NAME_var = tk.StringVar()
        self.DISTILLER_NAME_dropdown = ttk.Combobox(self.evaluation_config_frame, textvariable=self.DISTILLER_NAME_var, values=self.DISTILLER_NAME_options)
        self.DISTILLER_NAME_dropdown.grid(row=3, column=1, padx=10, pady=1)
        self.DISTILLER_NAME_dropdown.current(8)

        # Y_NAME
        self.Y_NAME_options = ["SCORES", "PORTFOLIOS"]
        tk.Label(self.evaluation_config_frame, text="Select Y Name:").grid(row=3, column=2, padx=10, pady=1)
        self.Y_NAME_var = tk.StringVar()
        self.Y_NAME_dropdown = ttk.Combobox(self.evaluation_config_frame, textvariable=self.Y_NAME_var, values=self.Y_NAME_options)
        self.Y_NAME_dropdown.grid(row=3, column=3, padx=10, pady=1)
        self.Y_NAME_dropdown.current(1)

        separator = ttk.Separator(self.evaluation_config_frame,  orient="horizontal")
        separator.grid(row=4, column=0, columnspan=4, padx=10, pady=1, sticky="ew")
        separator_label = ttk.Label(self.evaluation_config_frame, text="Metric Settings")
        separator_label.grid(row=4, column=0, columnspan=4, padx=10, pady=1)

        # TRANSACTIOS_COSTS_RATE
        tk.Label(self.evaluation_config_frame, text="Set Transactios Costs Rate:").grid(row=5, column=0, padx=10, pady=1)
        self.TRANSACTIOS_COSTS_RATE_var = tk.DoubleVar()
        self.TRANSACTIOS_COSTS_RATE_entry = tk.Entry(self.evaluation_config_frame, textvariable=self.TRANSACTIOS_COSTS_RATE_var)
        self.TRANSACTIOS_COSTS_RATE_entry.grid(row=5, column=1, padx=10, pady=1)
        self.TRANSACTIOS_COSTS_RATE_var.set(0.005)

        # TRANSACTIOS_COSTS_RATE_INTERVAL
        tk.Label(self.evaluation_config_frame, text="Set Transactios Costs Rate Interval:").grid(row=5, column=2, padx=10, pady=1)
        self.TRANSACTIOS_COSTS_RATE_INTERVAL_var = tk.DoubleVar()
        self.TRANSACTIOS_COSTS_RATE_INTERVAL_entry = tk.Entry(self.evaluation_config_frame, textvariable=self.TRANSACTIOS_COSTS_RATE_INTERVAL_var)
        self.TRANSACTIOS_COSTS_RATE_INTERVAL_entry.grid(row=5, column=3, padx=10, pady=1)
        self.TRANSACTIOS_COSTS_RATE_INTERVAL_var.set(0.001)


        self.evaluate_button = tk.Button(self.root, text="Evaluate Model", command=self.evaluate_model)
        self.evaluate_button.pack(padx=10, pady=5)

        # self.quit_button = tk.Button(self.root, text="Quit", command=self.root.quit)
        # self.quit_button.pack(padx=10, pady=1)

        # self.result_frame = tk.LabelFrame(self.root, text="Results")
        # self.result_frame.pack(padx=10, pady=1, fill="both", expand=True)

    def base_on_CRITERION_NAME(self, *args):
        # Enable or disable the text box based on the selected value
        selected_value = self.CRITERION_NAME_var.get()
        if selected_value == "LOG_WEALTH_L2_DIVERSIFICATION" or selected_value == "LOG_WEALTH_L2_CONCENTRATION":
            self.LAMBDA_L2_entry.config(state="normal")
        else:
            self.LAMBDA_L2_entry.config(state="disabled")

    def base_on_LOAD_LOCAL_DATALOADER(self):
        if self.LOAD_LOCAL_DATALOADER_var.get():
            self.SCALER_dropdown.config(state="disabled")
            self.INCLUDE_OHLCV_FEATURES_chk.config(state="disabled")
            self.INCLUDE_OVERLAP_FEATURES_chk.config(state="disabled")
            self.INCLUDE_MOMENTUM_FEATURES_chk.config(state="disabled")
            self.INCLUDE_VOLUME_FEATURES_chk.config(state="disabled")
            self.INCLUDE_CYCLE_FEATURES_chk.config(state="disabled")
            self.INCLUDE_PRICE_FEATURES_chk.config(state="disabled")
            self.INCLUDE_VOLATILITY_FEATURES_chk.config(state="disabled")
            self.INCLUDE_PATTERN_FEATURES_chk.config(state="disabled")
            self.INCLUDE_WINDOW_DATA_chk.config(state="disabled")
            self.WINDOW_SIZE_entry.config(state="disabled")
        else:
            self.SCALER_dropdown.config(state="normal")
            self.INCLUDE_OHLCV_FEATURES_chk.config(state="normal")
            self.INCLUDE_OVERLAP_FEATURES_chk.config(state="normal")
            self.INCLUDE_MOMENTUM_FEATURES_chk.config(state="normal")
            self.INCLUDE_VOLUME_FEATURES_chk.config(state="normal")
            self.INCLUDE_CYCLE_FEATURES_chk.config(state="normal")
            self.INCLUDE_PRICE_FEATURES_chk.config(state="normal")
            self.INCLUDE_VOLATILITY_FEATURES_chk.config(state="normal")
            self.INCLUDE_PATTERN_FEATURES_chk.config(state="normal")
            self.INCLUDE_WINDOW_DATA_chk.config(state="normal")
            self.WINDOW_SIZE_entry.config(state="normal")

    def base_on_TUNE_PARAMETERS(self):
        if self.TUNE_PARAMETERS_var.get():
            self.NUM_TRIALS_entry.config(state="normal")
        else:
            self.NUM_TRIALS_entry.config(state="disabled")

    def base_on_INCLUDE_WINDOW_DATA(self):
        if self.INCLUDE_WINDOW_DATA_var.get():
            self.WINDOW_SIZE_entry.config(state="normal")
        else:
            self.WINDOW_SIZE_entry.config(state="disabled")

    def load_dataset(self):
        config = load_config()

        config["GET_LATEST_FINOL"] = self.GET_LATEST_FINOL_var.get()
        config["DEVICE"] = self.DEVICE_var.get()
        config["MANUAL_SEED"] = self.MANUAL_SEED_var.get()

        config["LOAD_LOCAL_DATALOADER"] = self.LOAD_LOCAL_DATALOADER_var.get()
        config["DATASET_NAME"] = self.DATASET_NAME_var.get()
        config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_OHLCV_FEATURES"] = self.INCLUDE_OHLCV_FEATURES_var.get()
        config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_OVERLAP_FEATURES"] = self.INCLUDE_OVERLAP_FEATURES_var.get()
        config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_MOMENTUM_FEATURES"] = self.INCLUDE_MOMENTUM_FEATURES_var.get()
        config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_VOLUME_FEATURES"] = self.INCLUDE_VOLUME_FEATURES_var.get()
        config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_CYCLE_FEATURES"] = self.INCLUDE_CYCLE_FEATURES_var.get()
        config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_PRICE_FEATURES"] = self.INCLUDE_PRICE_FEATURES_var.get()
        config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_VOLATILITY_FEATURES"] = self.INCLUDE_VOLATILITY_FEATURES_var.get()
        config["FEATURE_ENGINEERING_CONFIG"]["INCLUDE_PATTERN_FEATURES"] = self.INCLUDE_PATTERN_FEATURES_var.get()
        config["DATA_AUGMENTATION_CONFIG"]["WINDOW_DATA"]["INCLUDE_WINDOW_DATA"] = self.INCLUDE_WINDOW_DATA_var.get()
        config["DATA_AUGMENTATION_CONFIG"]["WINDOW_DATA"]["WINDOW_SIZE"] = self.WINDOW_SIZE_var.get()
        config["SCALER"] = self.SCALER_var.get()

        update_config(config)

        try:
            self.load_dataset_output = DatasetLoader().load_dataset()
            messagebox.showinfo("Success", f"Dataset '{self.DATASET_NAME_var.get()}' loaded successfully!")
        except Exception as e:
            messagebox.showerror("Error", f"Failed to load dataset '{self.DATASET_NAME_var.get()}': {e}")


    def train_model(self):
        config = load_config()

        config["MODEL_NAME"] = self.MODEL_NAME_var.get()
        config["TUNE_PARAMETERS"] = self.TUNE_PARAMETERS_var.get()
        config["NUM_TRIALS"] = self.NUM_TRIALS_var.get()

        config["NUM_EPOCHES"] = self.NUM_EPOCHES_var.get()
        config["SAVE_EVERY"] = self.SAVE_EVERY_var.get()
        config["OPTIMIZER_NAME"] = self.OPTIMIZER_NAME_var.get()
        config["LEARNING_RATE"] = self.LEARNING_RATE_var.get()
        config["CRITERION_NAME"] = self.CRITERION_NAME_var.get()
        config["LAMBDA_L2"] = self.LAMBDA_L2_var.get()

        update_config(config)

        if hasattr(self, 'load_dataset_output'):
            try:
                self.train_model_output = ModelTrainer(self.load_dataset_output).train_model()
                messagebox.showinfo("Success", f"Model '{self.MODEL_NAME_var.get()}' trained successfully!")
            except Exception as e:
                messagebox.showerror("Error", f"Failed to train model '{self.MODEL_NAME_var.get()}': {e}")
        else:
            messagebox.showwarning("Warning", "Please load the dataset first!")

    def evaluate_model(self):
        config = load_config()

        config["PLOT_CHINESE"] = self.PLOT_CHINESE_var.get()
        config["PROP_WINNERS"] = self.PROP_WINNERS_var.get()
        config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_INTERPRETABILITY_ANALYSIS"] = self.INCLUDE_INTERPRETABILITY_ANALYSIS_var.get()
        config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_ECONOMIC_DISTILLATION"] = self.INCLUDE_ECONOMIC_DISTILLATION_var.get()
        config["INTERPRETABLE_ANALYSIS_CONFIG"]["PROP_DISTILLED_FEATURES"] = self.PROP_DISTILLED_FEATURES_var.get()
        config["INTERPRETABLE_ANALYSIS_CONFIG"]["DISTILLER_NAME"] = self.DISTILLER_NAME_var.get()
        config["INTERPRETABLE_ANALYSIS_CONFIG"]["Y_NAME"] = self.Y_NAME_var.get()
        config["METRIC_CONFIG"]["PRACTICAL_METRICS"]["TRANSACTIOS_COSTS_RATE"] = self.TRANSACTIOS_COSTS_RATE_var.get()
        config["METRIC_CONFIG"]["PRACTICAL_METRICS"]["TRANSACTIOS_COSTS_RATE_INTERVAL"] = self.TRANSACTIOS_COSTS_RATE_INTERVAL_var.get()

        update_config(config)

        if hasattr(self, 'load_dataset_output') and hasattr(self, 'train_model_output'):
            try:
                self.evaluate_model_output = ModelEvaluator(self.load_dataset_output, self.train_model_output).evaluate_model()
                messagebox.showinfo("Success", f"Model '{self.MODEL_NAME_var.get()}' evaluated successfully!")
            except Exception as e:
                messagebox.showerror("Error", f"Failed to evaluate model: '{self.MODEL_NAME_var.get()}': {e}")
        else:
            messagebox.showwarning("Warning", "Please load the dataset and train the model first!")


if __name__ == "__main__":
    root = tk.Tk()
    app = FinOLApp(root)
    root.mainloop()


import torch.nn as nn

from finol.data_layer.ScalerSelector import ScalerSelector
from finol.utils import load_config


class DNN(nn.Module):
    def __init__(self, model_args, model_params):
        super().__init__()
        self.config = load_config()

        self.layers = nn.ModuleList()
        self.layers.append(nn.Linear(model_args["NUM_FEATURES_AUGMENTED"], model_params["HIDDEN_SIZE"]))
        for _ in range(model_params["NUM_LAYERS"]):
            self.layers.append(nn.Linear(model_params["HIDDEN_SIZE"], model_params["HIDDEN_SIZE"]))
        self.layers.append(nn.Linear(model_params["HIDDEN_SIZE"], 1))
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=model_params["DROPOUT"])

    def forward(self, x):
        batch_size, num_assets, num_features_augmented = x.shape
        # x = x.view(batch_size, num_assets, window_size, num_features_original)

        """Input Transformation"""
        if self.config["SCALER"].startswith("Window"):
            x = ScalerSelector().window_normalize(x)

        out = x
        for layer in self.layers:
            out = layer(out)
            out = self.relu(out)
            out = self.dropout(out)

        """Final Scores for Assets"""
        final_scores = out.squeeze(-1)
        return final_scores


import os
import json
import time

ROOT_PATH = os.path.dirname(os.path.realpath(__file__))
PARENT_PATH = os.path.dirname(ROOT_PATH)

from rich import print

# with open(ROOT_PATH + "/config.json", "r") as f:
#     config = json.load(f)
#
# for key, value in config.items():
#     if not key.startswith('_'):  # 跳过注释键
#         globals()[key] = value

# print(f"GET_LATEST_FINOL: {globals()['GET_LATEST_FINOL']}")
# print(f"DATASET_NAME: {globals()['DATASET_NAME']}")

# GET_LATEST_FINOL = False
# TUTORIAL_MODE = True
# TUTORIAL_NAME = "TUTORIAL_4"
#
# ##########################################################################################
# ############################  Parameters related to data_layer ###########################
# ##########################################################################################
# LOAD_DATALOADER = True
# DATASET_NAME = "DJIA"  # Available options: NYSE(O), NYSE(N), DJIA, SP500, TSE, SSE, HSI, CMEG, CRYPTO, TUTORIAL, DJIA(N), Nasdaq-100
# DATASET_SPLIT_CONFIG = {
#     "NYSE(O)": {
#         "TRAIN_START_TIMESTAMP": "1962-07-03",
#         "TRAIN_END_TIMESTAMP": "1976-01-21",  # [3390 rows x 5 columns]
#         "VAL_START_TIMESTAMP": "1976-01-22",
#         "VAL_END_TIMESTAMP": "1980-07-11",  # [1130 rows x 5 columns]
#         "TEST_START_TIMESTAMP": "1980-07-14",
#         "TEST_END_TIMESTAMP": "1984-12-31"  # [1131 rows x 5 columns]
#     },
#     "NYSE(N)": {
#         "TRAIN_START_TIMESTAMP": "1985-01-02",
#         "TRAIN_END_TIMESTAMP": "2000-04-06",  # [3858 rows x 5 columns]
#         "VAL_START_TIMESTAMP": "2000-04-07",
#         "VAL_END_TIMESTAMP": "2005-05-20",  # [1286 rows x 5 columns]
#         "TEST_START_TIMESTAMP": "2005-05-23",
#         "TEST_END_TIMESTAMP": "2010-06-30"  # [1286 rows x 5 columns]
#     },
#     "DJIA": {
#         "TRAIN_START_TIMESTAMP": "2001-01-14",
#         "TRAIN_END_TIMESTAMP": "2002-04-01",  # [300 rows x 5 columns]
#         "VAL_START_TIMESTAMP": "2002-04-02",
#         "VAL_END_TIMESTAMP": "2002-08-21",  # [100 rows x 5 columns]
#         "TEST_START_TIMESTAMP": "2002-08-22",
#         "TEST_END_TIMESTAMP": "2003-01-14"  # [100 rows x 5 columns]
#     },
#     "SP500": {
#         "TRAIN_START_TIMESTAMP": "1998-01-02",
#         "TRAIN_END_TIMESTAMP": "2001-01-12",  # [756 rows x 5 columns]
#         "VAL_START_TIMESTAMP": "2001-01-16",
#         "VAL_END_TIMESTAMP": "2002-01-25",  # [256 rows x 5 columns]
#         "TEST_START_TIMESTAMP": "2002-01-28",
#         "TEST_END_TIMESTAMP": "2003-01-31"  # [256 rows x 5 columns]
#     },
#     "TSE": {
#         "TRAIN_START_TIMESTAMP": "1995-01-12",
#         "TRAIN_END_TIMESTAMP": "1997-05-28",  # [600 rows x 5 columns]
#         "VAL_START_TIMESTAMP": "1997-05-29",
#         "VAL_END_TIMESTAMP": "1998-03-13",  # [200 rows x 5 columns]
#         "TEST_START_TIMESTAMP": "1998-03-16",
#         "TEST_END_TIMESTAMP": "1998-12-31"  # [201 rows x 5 columns]
#     },
#     "SSE": {
#         "TRAIN_START_TIMESTAMP": "2010-07-05",
#         "TRAIN_END_TIMESTAMP": "2018-04-09",  # [406 rows x 5 columns]
#         "VAL_START_TIMESTAMP": "2018-04-16",
#         "VAL_END_TIMESTAMP": "2020-11-16",  # [136 rows x 5 columns]
#         "TEST_START_TIMESTAMP": "2020-11-23",
#         "TEST_END_TIMESTAMP": "2023-06-26"  # [136 rows x 5 columns]
#     },
#     "CRYPTO": {
#         "TRAIN_START_TIMESTAMP": "2017-11-09",
#         "TRAIN_END_TIMESTAMP": "2021-08-22",  # [1383 rows x 5 columns]
#         "VAL_START_TIMESTAMP": "2021-08-23",
#         "VAL_END_TIMESTAMP": "2022-11-26",  # [461 rows x 5 columns]
#         "TEST_START_TIMESTAMP": "2022-11-27",
#         "TEST_END_TIMESTAMP": "2024-03-01"  # [461 rows x 5 columns]
#     }
# }
# DATASET_SPLIT_CONFIG["HSI"] = DATASET_SPLIT_CONFIG["SSE"].copy()
# DATASET_SPLIT_CONFIG["CMEG"] = DATASET_SPLIT_CONFIG["SSE"].copy()
# DATASET_SPLIT_CONFIG["DJIA(N)"] = DATASET_SPLIT_CONFIG["SSE"].copy()
# DATASET_SPLIT_CONFIG["Nasdaq-100"] = DATASET_SPLIT_CONFIG["SSE"].copy()
#
# FEATURE_ENGINEERING_CONFIG = {
#     "INCLUDE_OHLCV_FEATURES": True,
#     "INCLUDE_OVERLAP_FEATURES": True,
#     "INCLUDE_MOMENTUM_FEATURES": True,
#     "INCLUDE_VOLUME_FEATURES": True,
#     "INCLUDE_CYCLE_FEATURES": True,
#     "INCLUDE_PRICE_FEATURES": True,
#     "INCLUDE_VOLATILITY_FEATURES": True,
#     "INCLUDE_PATTERN_FEATURES": True
# }
# DATA_AUGMENTATION_CONFIG = {
#     "WINDOW_DATA": {
#         "INCLUDE_WINDOW_DATA": True,
#         "WINDOW_SIZE": 50
#     }
# }
# SCALER = "WindowStandardScaler"  # None, StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler,
# # WindowStandardScaler, WindowMinMaxScaler, WindowMaxAbsScaler, WindowRobustScaler
# BATCH_SIZE = {
#     "NYSE(O)": 128,
#     "NYSE(N)": 128,
#     "DJIA": 64,
#     "SP500": 64,
#     "TSE": 64,
#     "SSE": 128,
#     "CRYPTO": 128,
# }
# BATCH_SIZE["HSI"] = BATCH_SIZE["SSE"]
# BATCH_SIZE["CMEG"] = BATCH_SIZE["SSE"]
# BATCH_SIZE["DJIA(N)"] = BATCH_SIZE["SSE"]
# BATCH_SIZE["Nasdaq-100"] = BATCH_SIZE["SSE"]
#
# DAYS = 52 if DATASET_NAME in ["SSE", "HSI", "CMEG", "DJIA(N)", 'Nasdaq-100'] else 252
#
# ##########################################################################################
# ########################### Parameters related to model_layer ############################
# ##########################################################################################
# MODEL_NAME = "LSRE-CAAN"  # AlphaPortfolio, CNN, DNN, LSRE-CAAN, LSTM, RNN, Transformer
# MODEL_CONFIG = {
#     "DNN": {
#         "NUM_LAYERS": 4,
#         "HIDDEN_SIZE": 32
#     },
#     "RNN": {
#         "NUM_LAYERS": 1,
#         "HIDDEN_SIZE": 32
#     },
#     "LSTM": {
#         "NUM_LAYERS": 4,
#         "HIDDEN_SIZE": 32
#     },
#     "CNN": {
#         "OUT_CHANNELS": 128,
#         "KERNEL_SIZE": 3,
#         "STRIDE": 1,
#         "HIDDEN_SIZE": 32,
#     },
#     "Transformer": {
#         "NUM_LAYERS": 1,
#         "NUM_HEADS": 1,
#         "HIDDEN_SIZE": 32,
#     },
#     "LSRE-CAAN": {
#         "NUM_LAYERS": 1,  # paper setting: 1
#         "NUM_LATENTS": 1,  # paper setting: 1
#         "LATENT_DIM": 32,  # paper setting: 32
#         "CROSS_HEADS": 1,  # paper setting: 1
#         "LATENT_HEADS": 1,  # paper setting: 1
#         "CROSS_DIM_HEAD": 64,  # paper setting: 64
#         "LATENT_DIM_HEAD": 32,  # paper setting: 32
#         "DROPOUT": 0.,  # paper setting: 0
#     },
#     "AlphaPortfolio": {
#         "DIM_EMBEDDING": 256,  # paper setting: 256
#         "DIM_FEEDFORWARD": 1024,  # paper setting: 1024
#         "NUM_HEADS": 4,  # paper setting: 4
#         "NUM_LAYERS": 1,  # paper setting: 1
#         "DROPOUT": 0.2,  # paper setting: 0.2
#     },
# }
# MODEL_CONFIG["LSRE-CAAN-d"] = MODEL_CONFIG["LSRE-CAAN"].copy()
# MODEL_CONFIG["LSRE-CAAN-dd"] = MODEL_CONFIG["LSRE-CAAN"].copy()
#
# MODEL_CONFIG_SPACE = {
#     'DNN': {
#         'NUM_LAYERS': {'type': 'int', 'range': (1, 12), 'step': 1},
#         'HIDDEN_SIZE': {'type': 'int', 'range': (32, 256), 'step': 32},
#         'DROPOUT': {'type': 'float', 'range': (0, 0.5), 'step': 0.05},
#     },
#     'Transformer': {
#         'NUM_LAYERS': {'type': 'int', 'range': (1, 12), 'step': 1},
#         'NUM_HEADS': {'type': 'int', 'range': (1, 12), 'step': 1},
#         'HIDDEN_SIZE': {'type': 'int', 'range': (32, 256), 'step': 32},
#         'DROPOUT': {'type': 'float', 'range': (0, 0.5), 'step': 0.05},
#     },
#     'LSRE-CAAN': {
#         'NUM_LAYERS': {'type': 'int', 'range': (1, 12), 'step': 1},
#         'NUM_LATENTS': {'type': 'int', 'range': (1, DATA_AUGMENTATION_CONFIG.get("WINDOW_DATA")["WINDOW_SIZE"]), 'step': 1},
#         'LATENT_DIM': {'type': 'int', 'range': (32, 256), 'step': 32},
#         'CROSS_HEADS': {'type': 'int', 'range': (1, 12), 'step': 1},
#         'LATENT_HEADS': {'type': 'int', 'range': (1, 12), 'step': 1},
#         'CROSS_DIM_HEAD': {'type': 'int', 'range': (32, 256), 'step': 32},
#         'LATENT_DIM_HEAD': {'type': 'int', 'range': (32, 256), 'step': 32},
#         'DROPOUT': {'type': 'float', 'range': (0, 0.5), 'step': 0.05},
#     },
#     "AlphaPortfolio": {
#         'DIM_EMBEDDING': {'type': 'int', 'range': (32, 256), 'step': 32},
#         'DIM_FEEDFORWARD': {'type': 'int', 'range': (32, 256), 'step': 32},
#         'NUM_HEADS': {'type': 'int', 'range': (1, 12), 'step': 1},
#         'NUM_LAYERS': {'type': 'int', 'range': (1, 12), 'step': 1},
#         'DROPOUT': {'type': 'float', 'range': (0, 0.5), 'step': 0.05},
#     },
# }
# MODEL_CONFIG_SPACE["RNN"] = MODEL_CONFIG_SPACE["DNN"].copy()
# MODEL_CONFIG_SPACE["LSTM"] = MODEL_CONFIG_SPACE["DNN"].copy()
# MODEL_CONFIG_SPACE["LSRE-CAAN-d"] = MODEL_CONFIG_SPACE["LSRE-CAAN"].copy()
# MODEL_CONFIG_SPACE["LSRE-CAAN-dd"] = MODEL_CONFIG_SPACE["LSRE-CAAN"].copy()
#
# DROPOUT = 0.1
# PROP_WINNERS = 0.5
#
# ##########################################################################################
# ######################## Parameters related to optimization_layer ########################
# ##########################################################################################
# TUNE_PARAMETERS = True
# MANUAL_SEED = 0  # 3442
# OPTIMIZER_NAME = "Lamb"
# LEARNING_RATE = 5e-4
# CRITERION_NAME = "LOG_WEALTH_L2_DIVERSIFICATION"  # LOG_WEALTH, LOG_WEALTH_L2_DIVERSIFICATION, LOG_WEALTH_L2_CONCENTRATION
# LAMBDA_L2 = 5e-4
# DEVICE = "cuda"
# NUM_EPOCHES = 1000
# SAVE_EVERY = 1
# PLOT_DYNAMIC_LOSS = True
#
#
# ##########################################################################################
# ######################### Parameters related to evaluation_layer #########################
# ##########################################################################################
# PLOT_CHINESE = False
# INTERPRETABLE_ANALYSIS_CONFIG = {
#     "INCLUDE_INTERPRETABILITY_ANALYSIS": False,
#     "INCLUDE_ECONOMIC_DISTILLATION": False,
#     "PROP_DISTILLED_FEATURES": 0.7,
#     "Y_NAME": "PORTFOLIOS",  # SCORES, PORTFOLIOS
#     "DISTILLER_NAME": "Lasso"
# }
# MARKERS = ['o', '^', '<', '>', 's', 'p', 'h', 'H', 'D', 'd']
# MARKEVERY = {
#     'NYSE(O)': 50,
#     'NYSE(N)': 55,
#     'DJIA': 3,
#     'SP500': 7,
#     'TSE': 6,
#     'SSE': 4,
#     'CRYPTO': 13,
#     }
# MARKEVERY["HSI"] = MARKEVERY["SSE"]
# MARKEVERY["CMEG"] = MARKEVERY["SSE"]
# MARKEVERY["DJIA(N)"] = MARKEVERY["SSE"]
# MARKEVERY["Nasdaq-100"] = MARKEVERY["SSE"]
# MARKEVERY = MARKEVERY.get(DATASET_NAME, 10)
#
# ALPHA = 0.5
# METRIC_CONFIG = {
#     "PRACTICAL_METRICS": {
#         "TRANSACTIOS_COSTS_RATE": 0.5/100,
#         "TRANSACTIOS_COSTS_RATE_INTERVAL": 1/1000,
#     }
# }
# BENCHMARK_BASELINE = ["Market", "Best", "UCRP", "BCRP"]
# FOLLOW_THE_WINNER = ["UP", "EG", "SCRP", "PPT", "SSPO"]
# FOLLOW_THE_LOSER = ["ANTI1", "ANTI2", "PAMR", "CWMR-Var", "CWMR-Stdev", "OLMAR-S", "OLMAR-E", "RMR", "RPRT"]
# PATTERN_MATCHING = ["AICTR", "KTPT"]
# META_LEARNING = ["SP", "ONS", "GRW", "WAAS", "CW-OGD"]
#
# # PLOT_ALL_1 = BENCHMARK_BASELINE + [MODEL_NAME]
# # PLOT_ALL_2 = FOLLOW_THE_WINNER + [MODEL_NAME]
# # PLOT_ALL_3 = FOLLOW_THE_LOSER + [MODEL_NAME]
# # PLOT_ALL_4 = PATTERN_MATCHING + [MODEL_NAME]
# # PLOT_ALL_5 = META_LEARNING + [MODEL_NAME]
#
# PLOT_ALL_1 = ["Market", "SCRP", "OLMAR-E", "RMR", "CW-OGD"] + [MODEL_NAME]
# # PLOT_ALL_2 = ["Market", "UP", "EG", "PAMR", "OLMAR-E", "RPRT", "KTPT", "CW-OGD"] + [MODEL_NAME]
# # PLOT_ALL_3 = ["Market", "UP", "EG", "PAMR", "OLMAR-E", "RPRT", "KTPT", "CW-OGD"] + [MODEL_NAME]
# # PLOT_ALL_4 = ["Market", "UP", "EG", "PAMR", "OLMAR-E", "RPRT", "KTPT", "CW-OGD"] + [MODEL_NAME]
# # PLOT_ALL_5 = ["Market", "UP", "EG", "PAMR", "OLMAR-E", "RPRT", "KTPT", "CW-OGD"] + [MODEL_NAME]
#
# if INTERPRETABLE_ANALYSIS_CONFIG['INCLUDE_ECONOMIC_DISTILLATION']:
#     PLOT_ALL_1 = PLOT_ALL_1 + [MODEL_NAME + ' (ED)']
#     # PLOT_ALL_2 = PLOT_ALL_2 + [MODEL_NAME + ' (ED)']
#     # PLOT_ALL_3 = PLOT_ALL_3 + [MODEL_NAME + ' (ED)']
#     # PLOT_ALL_4 = PLOT_ALL_4 + [MODEL_NAME + ' (ED)']
#     # PLOT_ALL_5 = PLOT_ALL_5 + [MODEL_NAME + ' (ED)']
#
# ################################################################################################################
# ############################################# FOR TUTORIAL ONLY ################################################
# ################################################################################################################
# if TUTORIAL_MODE:
#     LOAD_DATALOADER = True
#     DATASET_NAME = "DJIA"
#     FEATURE_ENGINEERING_CONFIG = {
#         "INCLUDE_OHLCV_FEATURES": True,
#         "INCLUDE_OVERLAP_FEATURES": True,
#         "INCLUDE_MOMENTUM_FEATURES": True,
#         "INCLUDE_VOLUME_FEATURES": True,
#         "INCLUDE_CYCLE_FEATURES": True,
#         "INCLUDE_PRICE_FEATURES": True,
#         "INCLUDE_VOLATILITY_FEATURES": True,
#         "INCLUDE_PATTERN_FEATURES": True
#     }
#     DATA_AUGMENTATION_CONFIG = {
#         "WINDOW_DATA": {
#             "INCLUDE_WINDOW_DATA": True,
#             "WINDOW_SIZE": 10
#         }
#     }
#     SCALER = "WindowStandardScaler"  # None, StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler,
#     # WindowStandardScaler, WindowMinMaxScaler, WindowMaxAbsScaler, WindowRobustScaler
#     BATCH_SIZE = {
#         "NYSE(O)": 128,
#         "NYSE(N)": 128,
#         "DJIA": 32,
#         "SP500": 64,
#         "TSE": 64,
#         "SSE": 32,
#         "HSI": 32,
#         "CMEG": 32,
#         "CRYPTO": 128,
#     }
#
#     # Parameters related to model_layer
#     MODEL_NAME = 'LSRE-CAAN'
#     MODEL_CONFIG = {
#         'DNN': {
#             'NUM_LAYERS': 1,
#             "HIDDEN_SIZE": 64,
#             "DROPOUT":  0.0
#         },
#         "RNN": {
#             "NUM_LAYERS": 1,
#             "HIDDEN_SIZE": 32,
#             "DROPOUT": 0.25,
#         },
#         "LSTM": {
#             "NUM_LAYERS": 4,
#             "HIDDEN_SIZE": 32,
#             "DROPOUT": 0.25,
#         },
#         "CNN": {
#             "OUT_CHANNELS": 128,
#             "KERNEL_SIZE": 3,
#             "STRIDE": 1,
#             "HIDDEN_SIZE": 32,
#             "DROPOUT": 0.25,
#         },
#         "Transformer": {
#             "NUM_LAYERS": 1,
#             "NUM_HEADS": 1,
#             "HIDDEN_SIZE": 32,
#             "DROPOUT": 0.25,
#         },
#         "LSRE-CAAN": {
#             "NUM_LAYERS": 1,  # paper setting: 1
#             "NUM_LATENTS": 1,  # paper setting: 1
#             "LATENT_DIM": 32,  # paper setting: 32
#             "CROSS_HEADS": 1,  # paper setting: 1
#             "LATENT_HEADS": 1,  # paper setting: 1
#             "CROSS_DIM_HEAD": 64,  # paper setting: 64
#             "LATENT_DIM_HEAD": 32,  # paper setting: 32
#             "DROPOUT": 0.,  # paper setting: 0
#         },
#         "AlphaPortfolio": {
#             "DIM_EMBEDDING": 256,  # paper setting: 256
#             "DIM_FEEDFORWARD": 1024,  # paper setting: 1024
#             "NUM_HEADS": 4,  # paper setting: 4
#             "NUM_LAYERS": 1,  # paper setting: 1
#             "DROPOUT": 0.2,  # paper setting: 0.2
#         },
#     }
#     MODEL_CONFIG_SPACE = {
#         'DNN': {
#             'NUM_LAYERS': {'type': 'int', 'range': (1, 12), 'step': 1},
#             'HIDDEN_SIZE': {'type': 'int', 'range': (32, 256), 'step': 32},
#             'DROPOUT': {'type': 'float', 'range': (0, 0.5), 'step': 0.05},
#         },
#         'Transformer': {
#             'NUM_LAYERS': {'type': 'int', 'range': (1, 12), 'step': 1},
#             'NUM_HEADS': {'type': 'int', 'range': (1, 12), 'step': 1},
#             'HIDDEN_SIZE': {'type': 'int', 'range': (32, 256), 'step': 32},
#             'DROPOUT': {'type': 'float', 'range': (0, 0.5), 'step': 0.05},
#         },
#         'LSRE-CAAN': {
#             'NUM_LAYERS': {'type': 'int', 'range': (1, 12), 'step': 1},
#             'NUM_LATENTS': {'type': 'int', 'range': (1, DATA_AUGMENTATION_CONFIG.get("WINDOW_DATA")["WINDOW_SIZE"]), 'step': 1},
#             'LATENT_DIM': {'type': 'int', 'range': (32, 256), 'step': 32},
#             'CROSS_HEADS': {'type': 'int', 'range': (1, 12), 'step': 1},
#             'LATENT_HEADS': {'type': 'int', 'range': (1, 12), 'step': 1},
#             'CROSS_DIM_HEAD': {'type': 'int', 'range': (32, 256), 'step': 32},
#             'LATENT_DIM_HEAD': {'type': 'int', 'range': (32, 256), 'step': 32},
#             'DROPOUT': {'type': 'float', 'range': (0, 0.5), 'step': 0.05},
#         },
#         "AlphaPortfolio": {
#             'DIM_EMBEDDING': {'type': 'int', 'range': (32, 256), 'step': 32},
#             'DIM_FEEDFORWARD': {'type': 'int', 'range': (32, 256), 'step': 32},
#             'NUM_HEADS': {'type': 'int', 'range': (1, 12), 'step': 1},
#             'NUM_LAYERS': {'type': 'int', 'range': (1, 12), 'step': 1},
#             'DROPOUT': {'type': 'float', 'range': (0, 0.5), 'step': 0.05},
#         },
#     }
#     MODEL_CONFIG_SPACE["RNN"] = MODEL_CONFIG_SPACE["DNN"].copy()
#     MODEL_CONFIG_SPACE["LSTM"] = MODEL_CONFIG_SPACE["DNN"].copy()
#     MODEL_CONFIG_SPACE["LSRE-CAAN-d"] = MODEL_CONFIG_SPACE["LSRE-CAAN"].copy()
#     MODEL_CONFIG_SPACE["LSRE-CAAN-dd"] = MODEL_CONFIG_SPACE["LSRE-CAAN"].copy()
#     PROP_WINNERS = 1
#
#     # Parameters related to optimization_layer
#     TUNE_PARAMETERS = True
#     MANUAL_SEED = 0  # 3442
#     OPTIMIZER_NAME = "Adam"
#     LEARNING_RATE = 5e-4
#     CRITERION_NAME = "LOG_WEALTH"  # LOG_WEALTH, LOG_WEALTH_L2_DIVERSIFICATION, LOG_WEALTH_L2_CONCENTRATION
#     LAMBDA_L2 = 5e-4
#     DEVICE = "cuda"
#     NUM_EPOCHES = 100
#     SAVE_EVERY = 1
#     PLOT_DYNAMIC_LOSS = True
#
#     # Parameters related to evaluation_layer
#     METRIC_CONFIG = {
#         "INCLUDE_PROFIT_METRICS": True,
#         "INCLUDE_RISK_METRICS": True,
#         "PRACTICAL_METRICS": {
#             "INCLUDE_PRACTICAL_METRICS": True,
#             "TRANSACTIOS_COSTS_RATE": 1 / 100,
#             "TRANSACTIOS_COSTS_RATE_INTERVAL": 1 / 1000,
#         }
#     }
#
#     PLOT_ALL_1 = ["Market", "SCRP", "OLMAR-E", "RMR", "CW-OGD"] + [MODEL_NAME]
#     # PLOT_ALL_2 = ["Market", "UP", "EG", "PAMR", "OLMAR-E", "RPRT", "KTPT", "CW-OGD"] + [MODEL_NAME]
#     # PLOT_ALL_3 = ["Market", "UP", "EG", "PAMR", "OLMAR-E", "RPRT", "KTPT", "CW-OGD"] + [MODEL_NAME]
#     # PLOT_ALL_4 = ["Market", "UP", "EG", "PAMR", "OLMAR-E", "RPRT", "KTPT", "CW-OGD"] + [MODEL_NAME]
#     # PLOT_ALL_5 = ["Market", "UP", "EG", "PAMR", "OLMAR-E", "RPRT", "KTPT", "CW-OGD"] + [MODEL_NAME]
#
#     if INTERPRETABLE_ANALYSIS_CONFIG['INCLUDE_ECONOMIC_DISTILLATION']:
#         PLOT_ALL_1 = PLOT_ALL_1 + [MODEL_NAME + ' (ED)']
#
#     if TUTORIAL_NAME == "TUTORIAL_1":
#         DATASET_NAME = TUTORIAL_NAME
#         DATASET_SPLIT_CONFIG = {
#             DATASET_NAME: {
#                 "TRAIN_START_TIMESTAMP": "2023-05-23",
#                 "TRAIN_END_TIMESTAMP": "2023-10-24",
#                 "VAL_START_TIMESTAMP": "2023-10-25",
#                 "VAL_END_TIMESTAMP": "2023-11-26",
#                 "TEST_START_TIMESTAMP": "2023-11-27",
#                 "TEST_END_TIMESTAMP": "2023-12-28"
#             }
#         }



import sys
import subprocess
import os
import time
import json
import requests
import torch
import random
import numpy as np
import torch.nn.functional as F

from shutil import copy2
from finol.update import get_latest_version
from finol import __version__


ROOT_PATH = os.path.dirname(os.path.realpath(__file__))
PARENT_PATH = os.path.dirname(ROOT_PATH)


def check_update():
    config = load_config()
    GET_LATEST_FINOL = config["GET_LATEST_FINOL"]

    latest = get_latest_version()
    if __version__ == latest:
        print("The current FinOL is latest")
    else:
        print("The current FinOL is not latest, please update by `pip install --upgrade finol`")
        print("Before updating, remember to back up any modifications you made to the FinOL project, such as added model code.")
        print("Note that `pip install --upgrade finol` will overwrite all files except the `logdir` folder, so you don't need to back up the `logdir`.")

        if GET_LATEST_FINOL:
            sys.exit()


def make_logdir():
    logdir = PARENT_PATH + "/logdir/" + str(time.strftime("%Y-%m-%d_%H-%M-%S", time.localtime()))
    os.makedirs(logdir)
    copy2(ROOT_PATH + "/config.json", logdir)

    return logdir


def load_config():
    with open(ROOT_PATH + "/config.json", "r") as f:
        config = json.load(f)
    return config


def update_config(config):
    with open(ROOT_PATH + "/config.json", "w") as f:
        json.dump(config, f, indent=4)


def download_data():
    print("data downloading...")
    github_url = "https://github.com/ai4finol/finol_data.git"
    local_path = ROOT_PATH + r"\data"
    subprocess.run(["git", "clone", github_url, local_path])

def portfolio_selection(final_scores):
    portfolio = F.softmax(final_scores, dim=-1)
    # print(final_scores)
    # print(portfolio)
    assert torch.all(portfolio >= 0), "Portfolio contains non-negative values."
    return portfolio


def actual_portfolio_selection(final_scores):
    config = load_config()
    PROP_WINNERS = config["PROP_WINNERS"]
    DEVICE = config["DEVICE"]

    NUM_ASSETS = final_scores.shape[-1]
    if PROP_WINNERS == 1:
        portfolio = F.softmax(final_scores, dim=-1)
    else:
        NUM_SELECTED_ASSETS = int(NUM_ASSETS * PROP_WINNERS)
        assert NUM_SELECTED_ASSETS > 0
        assert NUM_SELECTED_ASSETS <= NUM_ASSETS

        values, indices = torch.topk(final_scores, k=NUM_SELECTED_ASSETS)
        winners_mask = torch.ones_like(final_scores, device=DEVICE)
        winners_mask.scatter_(1, indices, 0).detach()
        portfolio = F.softmax(final_scores - 1e9 * winners_mask, dim=-1)
    assert torch.all(portfolio >= 0), "Portfolio contains non-negative values."
    return portfolio


def send_message_dingding(dingding_message):
    headers_dingding = {
        "Content-Type": "application/json",
        "Charset": "UTF-8"
    }
    content = f"\n" \
              f"message：\tmodel finish training & testing\n" \
              f"logdir：\t{dingding_message['logdir']}\n" \
              f"CW：\t{dingding_message['CW']}\n" \

    webhook = "https://oapi.dingtalk.com/robot/send?access_token=xxxxxxxxxxxxxxxxxxxxxxxx"  # your own dingding api
    message = {
        "msgtype": "text",
        "text": {
            "content": content
        },
        "at": {
            "isAtAll": True
        }
    }
    message_json = json.dumps(message)
    try:
        info = requests.post(url=webhook, data=message_json, headers=headers_dingding, verify=False).json()
    except Exception as e:
        print(f"{e}")
        return
    if info.get("errcode") != 0:
        print(f"{info}")


def get_variable_name(var):
    return next(key for key, value in globals().items() if value is var)


def set_seed(seed):
    os.environ["PYTHONHASHSEED"] = str(seed)
    # os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True


import torch.nn as nn

from einops import rearrange
from finol.data_layer.ScalerSelector import ScalerSelector
from finol.utils import load_config


class CNN(nn.Module):
    def __init__(self, model_args, model_params):
        super().__init__()
        self.config = load_config()
        self.model_args = model_args
        self.model_params = model_params

        self.net = nn.Sequential(
            nn.Conv1d(model_args["NUM_FEATURES_ORIGINAL"], model_params["OUT_CHANNELS"], model_params["KERNEL_SIZE"], model_params["STRIDE"]),
            nn.ReLU(),
            nn.AdaptiveAvgPool1d(1),
            nn.Flatten(),
            nn.Linear(model_params["OUT_CHANNELS"], model_params["HIDDEN_SIZE"]),
            nn.ReLU(),
            nn.Linear(model_params["HIDDEN_SIZE"], 1),
        )

    def forward(self, x):
        batch_size, num_assets, num_features_augmented = x.shape

        """Input Transformation"""
        x = x.view(batch_size, num_assets, self.model_args["WINDOW_SIZE"], self.model_args["NUM_FEATURES_ORIGINAL"])
        x = rearrange(x, "b m n d -> (b m) n d")
        x = x.transpose(1, 2)  # [batch_size * num_assets, seq_len, num_inputs] -> [batch_size * num_assets, num_inputs, seq_len]

        if self.config["SCALER"].startswith("Window"):
            x = ScalerSelector().window_normalize(x)

        """Temporal Representation Extraction"""
        out = self.net(x)
        out = out.view(batch_size, num_assets, 1).squeeze(-1)

        """Final Scores for Assets"""
        final_scores = out
        return final_scores



import warnings
import pandas as pd
import matplotlib.pyplot as plt

from finol.evaluation_layer.radar_chart import plot_radar_chart
from finol.utils import ROOT_PATH, load_config

plt.style.use("seaborn-paper")
plt.rcParams["font.family"] = "Microsoft YaHei"
warnings.simplefilter(action="ignore", category=FutureWarning)


class BenchmarkLoader:
    def __init__(self, caculate_metric_output, economic_distiller_caculate_metric_output):
        self.config = load_config()
        self.caculate_metric_output = caculate_metric_output
        self.economic_distiller_caculate_metric_output = economic_distiller_caculate_metric_output
        self.logdir = self.caculate_metric_output["logdir"]

    def plot_dataframe(self, df, plot_type):
        if plot_type == "DCW":
            df_copy = df.copy()
            df_copy = df_copy.drop(df.columns[-1], axis=1)
            df_copy.set_index('DATE', inplace=True)
            last_row = df_copy.iloc[-1]
            max_columns = last_row.nlargest(5).index.tolist()
            column_names = max_columns + [self.config["MODEL_NAME"]]
            print(
                column_names
            )
        else:
            column_names = self.config["COMPARED_BASELINE"] + [self.config["MODEL_NAME"]]

        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_ECONOMIC_DISTILLATION"]:
            column_names = column_names + [self.config["MODEL_NAME"] + ' (ED)']

        fig = plt.figure()  # figsize=(9, 4)

        if plot_type == "DCW":
            markevery = self.config["MARKEVERY"][self.config["DATASET_NAME"]]
            if self.config["PLOT_CHINESE"]:
                xlabel = "交易期"
                ylabel = "逐期累积财富"
            else:
                xlabel = "Trading Periods"
                ylabel = "Daily Cumulative Wealth"

        elif plot_type == "DDD":
            markevery = self.config["MARKEVERY"][self.config["DATASET_NAME"]]
            if self.config["PLOT_CHINESE"]:
                xlabel = "交易期"
                ylabel = "逐期下行风险"
            else:
                xlabel = "Trading Periods"
                ylabel = "Daily DrawDown"

        elif plot_type == "TCW":
            plt.figure()  # figsize=(6, 5)
            # TRANSACTIOS_COSTS_RATE = self.config["METRIC_CONFIG"].get("PRACTICAL_METRICS")["TRANSACTIOS_COSTS_RATE"]
            # TRANSACTIOS_COSTS_RATE_INTERVAL = self.config["METRIC_CONFIG"].get("PRACTICAL_METRICS")["TRANSACTIOS_COSTS_RATE_INTERVAL"]
            TRANSACTIOS_COSTS_RATE = self.config["METRIC_CONFIG"]["PRACTICAL_METRICS"]["TRANSACTIOS_COSTS_RATE"]
            TRANSACTIOS_COSTS_RATE_INTERVAL = self.config["METRIC_CONFIG"]["PRACTICAL_METRICS"]["TRANSACTIOS_COSTS_RATE_INTERVAL"]
            markevery = 1
            if self.config["PLOT_CHINESE"]:
                xlabel = "交易费用率 (%)"
                ylabel = "考虑交易费用的累积财富"
            else:
                xlabel = "Transaction Costs Rates (%)"
                ylabel = "Costs-Adjusted Cumulative Wealth"

            plt.xticks(ticks=[0, 2, 4, 6, 8, 10], labels=[0, 0.2, 0.4, 0.6, 0.8, 1])
            plt.xticks(ticks=[0, 2.5, 5, 7.5, 10], labels=[0, 0.25, 0.5, 0.75, 1])
            plt.xticks(ticks=[0, 1.25, 2.5, 3.75, 5, 7.5, 10], labels=[0, 0.125, 0.25, 0.375, 0.5, 0.75, 1])

        df = df[column_names]
        # df = df.rename(columns={df.columns[-2]: "Teacher-Model", df.columns[-1]: "Student-Model"})
        num_columns = len(df.columns)

        if self.config["INTERPRETABLE_ANALYSIS_CONFIG"]["INCLUDE_ECONOMIC_DISTILLATION"]:
            colors = ["black"] * (num_columns - 2) + ["red"] * 2
            lines = ["-"] * (num_columns - 1) + [":"]
        else:
            colors = ["black"] * (num_columns - 1) + ["red"] * 1
            lines = ["-"] * (num_columns)

        if plot_type == "TCW":
            for i, column in enumerate(df.columns):
                plt.plot(df[column].head(int(TRANSACTIOS_COSTS_RATE/TRANSACTIOS_COSTS_RATE_INTERVAL)+1), linestyle=lines[i], color=colors[i], marker=self.config["MARKERS"][i], markevery=markevery, alpha=0.5,
                         label=column)
        else:
            for i, column in enumerate(df.columns):
                plt.plot(df[column], linestyle=lines[i], color=colors[i], marker=self.config["MARKERS"][i], markevery=markevery, alpha=0.5,
                         label=column)

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(self.config["DATASET_NAME"])
        plt.legend(loc="best")  # lower left  upper left
        plt.grid(True)
        # plt.yscale("log")
        plt.tight_layout()
        plt.savefig(self.logdir + "/" + self.config["MODEL_NAME"] + "_" + self.config["DATASET_NAME"] + "_" + plot_type + ".pdf",
                    format="pdf",
                    dpi=300,
                    bbox_inches="tight")
        plt.show()

    def load_benchmark(self):
        """
        Load baseline results from a CSV file for specified target classes.

        Returns:
        - None
        """
        if self.caculate_metric_output != None:
            logdir = self.caculate_metric_output["logdir"]
        else:
            logdir = ROOT_PATH

        # if METRIC_CONFIG.get("INCLUDE_PROFIT_METRICS"):
        daily_return = pd.read_excel(ROOT_PATH + "/data/benchmark_results/profit_metrics/" + self.config["DATASET_NAME"] + "/daily_return.xlsx")
        daily_cumulative_wealth = pd.read_excel(ROOT_PATH + "/data/benchmark_results/profit_metrics/" + self.config["DATASET_NAME"] + "/daily_cumulative_wealth.xlsx")
        final_profit_result = pd.read_excel(ROOT_PATH + "/data/benchmark_results/profit_metrics/" + self.config["DATASET_NAME"] + "/final_profit_result.xlsx")

        daily_return = daily_return.dropna(axis=1, how="any")
        daily_cumulative_wealth = daily_cumulative_wealth.dropna(axis=1, how="any")
        final_profit_result = final_profit_result.dropna(axis=1, how="any")
        if self.caculate_metric_output != None:
            daily_cumulative_wealth[self.config["MODEL_NAME"]] = self.caculate_metric_output["DCW"]
            final_profit_result = final_profit_result.assign(**{self.config["MODEL_NAME"]: 0})
            final_profit_result.loc[0, self.config["MODEL_NAME"]] = self.caculate_metric_output["CW"]
            final_profit_result.loc[1, self.config["MODEL_NAME"]] = self.caculate_metric_output["APY"]
            final_profit_result.loc[2, self.config["MODEL_NAME"]] = self.caculate_metric_output["SR"]
        if self.economic_distiller_caculate_metric_output != None:
            daily_cumulative_wealth[self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["DCW"]
            final_profit_result = final_profit_result.assign(**{self.config["MODEL_NAME"] + " (ED)": 0})
            final_profit_result.loc[0, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["CW"]
            final_profit_result.loc[1, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["APY"]
            final_profit_result.loc[2, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["SR"]
        print(
            final_profit_result
        )
        # dd = final_profit_result.drop("Metric", axis=1).T
        # dd.columns = ["CW", "APY", "SR"]
        # dd = dd.reset_index(drop=False)
        # print(
        #     dd
        # )
        # # time.sleep(1111)
        # import dash
        # from dash import dash_table
        # from dash import html
        # # Create Dash app
        # app = dash.Dash(__name__)
        # # Create Dash table from dataframe
        # table = dash_table.DataTable(
        #   data=dd.to_dict("records"),
        #   columns=[{"name": i, "id": i} for i in dd.columns]
        # )
        # # Add sorting
        # table.sort_action = "native"
        # app.layout = html.Div([
        #     html.H1("Interactive DataTable"),
        #
        #     html.Div([
        #         html.H2("DataFrame 1"),
        #         table
        #     ]),
        #
        #     html.Div([
        #         html.H3("DataFrame 2"),
        #         table
        #     ]),
        # ])
        # app.run_server(debug=True)
        # time.sleep(1111)

        self.plot_dataframe(daily_cumulative_wealth, "DCW")
        # plot_dataframe(daily_cumulative_wealth, PLOT_ALL_2, "DCW", logdir)
        # plot_dataframe(daily_cumulative_wealth, PLOT_ALL_3, "DCW", logdir)
        # plot_dataframe(daily_cumulative_wealth, PLOT_ALL_4, "DCW", logdir)
        # plot_dataframe(daily_cumulative_wealth, PLOT_ALL_5, "DCW", logdir)

        # if METRIC_CONFIG.get("INCLUDE_RISK_METRICS"):
        daily_drawdown = pd.read_excel(ROOT_PATH + "/data/benchmark_results/risk_metrics/" + self.config["DATASET_NAME"] + "/daily_drawdown.xlsx")
        final_risk_result = pd.read_excel(ROOT_PATH + "/data/benchmark_results/risk_metrics/" + self.config["DATASET_NAME"] + "/final_risk_result.xlsx")

        daily_drawdown = daily_drawdown.dropna(axis=1, how="any")
        final_risk_result = final_risk_result.dropna(axis=1, how="any")
        if self.caculate_metric_output != None:
            daily_drawdown[self.config["MODEL_NAME"]] = self.caculate_metric_output["DDD"]
            final_risk_result = final_risk_result.assign(**{self.config["MODEL_NAME"]: 0})
            final_risk_result.loc[0, self.config["MODEL_NAME"]] = self.caculate_metric_output["VR"]
            final_risk_result.loc[1, self.config["MODEL_NAME"]] = self.caculate_metric_output["MDD"]
        if self.economic_distiller_caculate_metric_output != None:
            daily_drawdown[self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["DDD"]
            final_risk_result = final_risk_result.assign(**{self.config["MODEL_NAME"] + " (ED)": 0})
            final_risk_result.loc[0, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["VR"]
            final_risk_result.loc[1, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["MDD"]
        print(
            final_risk_result
        )
        self.plot_dataframe(daily_drawdown, "DDD")
        # plot_dataframe(daily_drawdown, PLOT_ALL_2, "DDD", logdir)
        # plot_dataframe(daily_drawdown, PLOT_ALL_3, "DDD", logdir)
        # plot_dataframe(daily_drawdown, PLOT_ALL_4, "DDD", logdir)
        # plot_dataframe(daily_drawdown, PLOT_ALL_5, "DDD", logdir)

        # if METRIC_CONFIG.get("PRACTICAL_METRICS")["INCLUDE_PRACTICAL_METRICS"]:
        transaction_costs_adjusted_cumulative_wealth = pd.read_excel(ROOT_PATH + "/data/benchmark_results/practical_metrics/" + self.config["DATASET_NAME"] + "/transaction_costs_adjusted_cumulative_wealth.xlsx")
        final_practical_result = pd.read_excel(ROOT_PATH + "/data/benchmark_results/practical_metrics/" + self.config["DATASET_NAME"] + "/final_practical_result.xlsx")

        transaction_costs_adjusted_cumulative_wealth = transaction_costs_adjusted_cumulative_wealth.dropna(axis=1, how="any")
        final_practical_result = final_practical_result.dropna(axis=1, how="any")
        if self.caculate_metric_output != None:
            transaction_costs_adjusted_cumulative_wealth[self.config["MODEL_NAME"]] = self.caculate_metric_output["TCW"]
            final_practical_result = final_practical_result.assign(**{self.config["MODEL_NAME"]: 0})
            final_practical_result.loc[0, self.config["MODEL_NAME"]] = self.caculate_metric_output["ATO"]
            final_practical_result.loc[1, self.config["MODEL_NAME"]] = self.caculate_metric_output["RT"]
        if self.economic_distiller_caculate_metric_output != None:
            transaction_costs_adjusted_cumulative_wealth[self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["TCW"]
            final_practical_result = final_practical_result.assign(**{self.config["MODEL_NAME"] + " (ED)": 0})
            final_practical_result.loc[0, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["ATO"]
            final_practical_result.loc[1, self.config["MODEL_NAME"] + " (ED)"] = self.economic_distiller_caculate_metric_output["RT"]
        print(
            final_practical_result
        )
        self.plot_dataframe(transaction_costs_adjusted_cumulative_wealth, "TCW")
        # plot_dataframe(transaction_costs_adjusted_cumulative_wealth, PLOT_ALL_2, "TCW", logdir)
        # plot_dataframe(transaction_costs_adjusted_cumulative_wealth, PLOT_ALL_3, "TCW", logdir)
        # plot_dataframe(transaction_costs_adjusted_cumulative_wealth, PLOT_ALL_4, "TCW", logdir)
        # plot_dataframe(transaction_costs_adjusted_cumulative_wealth, PLOT_ALL_5, "TCW", logdir)

        plot_radar_chart(final_profit_result, final_risk_result, self.COMPARED_BASELINE, logdir)

        load_benchmark_output = {}
        load_benchmark_output["logdir"] = self.caculate_metric_output["logdir"]
        load_benchmark_output["CW"] = self.caculate_metric_output["CW"]
        load_benchmark_output["TCW"] = self.caculate_metric_output["TCW"].iloc[-6]
        print(load_benchmark_output["TCW"])
        return load_benchmark_output


# if __name__ == "__main__":



import torch
import torch.nn.functional as F

from torch import nn
from einops import rearrange
from finol.data_layer.ScalerSelector import ScalerSelector
from finol.utils import load_config

"""
     Table C.1: Hyperparameters of TE-CAAN-Based AP
+----------------------+--------+-----------------+--------+
| Hyper-parameter      | Choice | Hyper-parameter | Choice |
+----------------------+--------+-----------------+--------+
| Embedding dimension  | 256    | Optimizer       | SGD    |
| Feed-forward network | 1021   | Learning rate   | 0.0001 |
| Number of multi-head | 4      | Dropout ratio   | 0.2    |
| Number of TE layer   | 1      | Training epochs | 30     |
+----------------------+--------+-----------------+--------+
"""


class SREM(nn.Module):
    """
    This class implements the Sequence Representations Extraction (SREM) module

    For more details, please refer to the papers `AlphaPortfolio: Direct Construction through Reinforcement Learning
    and Interpretable AI <https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3698800>` and `Attention is all you need
    <https://proceedings.neurips.cc/paper_files/paper/2017/hash/3f5ee243547dee91fbd053c1c4a845aa-Abstract.html>`
    """
    def __init__(self, model_args, model_params):
        super().__init__()
        self.token_emb = nn.Linear(model_args["NUM_FEATURES_ORIGINAL"], model_params["DIM_EMBEDDING"])
        self.pos_emb = nn.Embedding(model_args["WINDOW_SIZE"], model_params["DIM_EMBEDDING"])
        self.transformer_encoder = nn.TransformerEncoder(
            nn.TransformerEncoderLayer(
                d_model=model_params["DIM_EMBEDDING"],
                nhead=model_params["NUM_HEADS"],
                dim_feedforward=model_params["DIM_FEEDFORWARD"],
                dropout=model_params["DROPOUT"],
                batch_first=True,
            ),
            num_layers=model_params["NUM_LAYERS"],
        )

    def forward(self, x):
        """
        Args:
            x (Tensor): the sequence to the encoder (required).
                        shape (batch_size * num_assets, window_size, num_features_original)
        """
        _, n, d, device = x.shape[0], x.shape[1], x.shape[2], x.device  # n: window size; d: number of features
        x = self.token_emb(x)  # [batch_size * num_assets, window_size, num_features_original] -> [batch_size * num_assets, window_size, DIM_EMBEDDING]
        pos_emb = self.pos_emb(torch.arange(n, device=device))
        pos_emb = rearrange(pos_emb, "n d -> () n d")
        x = x + pos_emb

        x = self.transformer_encoder(x)  # [batch_size * num_assets, window_size, DIM_EMBEDDING] -> [batch_size * num_assets, window_size, DIM_EMBEDDING]

        return torch.mean(x, dim=1)  # [batch_size * num_assets, window_size, DIM_EMBEDDING] -> [batch_size * num_assets, DIM_EMBEDDING]


class CAAN(nn.Module):
    """
    This class implements the Cross Asset Attention Network (CAAN) module
    """
    def __init__(self, model_params):
        super().__init__()
        self.linear_query = torch.nn.Linear(model_params["DIM_EMBEDDING"], model_params["DIM_EMBEDDING"])
        self.linear_key = torch.nn.Linear(model_params["DIM_EMBEDDING"], model_params["DIM_EMBEDDING"])
        self.linear_value = torch.nn.Linear(model_params["DIM_EMBEDDING"], model_params["DIM_EMBEDDING"])
        self.linear_winner = torch.nn.Linear(model_params["DIM_EMBEDDING"], 1)

    def forward(self, x):
        query = self.linear_query(x)  # [batch_size, num_assets, DIM_EMBEDDING]
        key = self.linear_key(x)  # [batch_size, num_assets, DIM_EMBEDDING]
        value = self.linear_value(x)  # [batch_size, num_assets, DIM_EMBEDDING]

        beta = torch.matmul(query, key.transpose(1, 2)) / torch.sqrt(torch.tensor(float(query.shape[-1])))  # [batch_size, num_assets, DIM_EMBEDDING]
        beta = F.softmax(beta, dim=-1).unsqueeze(-1)
        x = torch.sum(value.unsqueeze(1) * beta, dim=2)  # [batch_size, num_assets, DIM_EMBEDDING]

        final_scores = self.linear_winner(x).squeeze(-1)  # [batch_size, num_assets]

        return final_scores


class AlphaPortfolio(nn.Module):
    """
    This class implements the AlphaPortfolio model

    For more details, please refer to the paper `AlphaPortfolio: Direct Construction through Reinforcement Learning
    and Interpretable AI <https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3698800>`
    """
    def __init__(self, model_args, model_params):
        super().__init__()
        self.config = load_config()
        self.model_args = model_args
        self.model_params = model_params

        self.srem = SREM(model_args, model_params)
        self.caan = CAAN(model_params)

    def forward(self, x):
        batch_size, num_assets, num_features_augmented = x.shape

        """Input Transformation"""
        x = x.view(batch_size, num_assets, self.model_args["WINDOW_SIZE"], self.model_args["NUM_FEATURES_ORIGINAL"])
        x = rearrange(x, "b m n d -> (b m) n d")
        if self.config["SCALER"].startswith("Window"):
            x = ScalerSelector().window_normalize(x)

        """Sequence Representations Extraction (SREM)"""
        stock_rep = self.srem(x)  # [batch_size * num_assets, window_size, DIM_EMBEDDING] -> [batch_size * num_assets, DIM_EMBEDDING]
        x = stock_rep.view(batch_size, num_assets, self.model_params["DIM_EMBEDDING"])  # [batch_size * num_assets, DIM_EMBEDDING] -> [batch_size, num_assets, DIM_EMBEDDING]

        """Cross Asset Attention Network (CAAN)"""
        final_scores = self.caan(x)

        return final_scores


if __name__ == "__main__":
    # config = load_config()
    DEVICE = "cuda"
    torch.manual_seed(0)
    batch_size = 128
    num_assets = 6
    window_size = 30
    num_features_original = 10
    num_features_augmented = window_size * num_features_original
    # x = torch.ones(batch_size, num_assets, num_features_augmented).to(DEVICE)
    x = torch.rand(batch_size, num_assets, num_features_augmented).to(DEVICE)
    model_args = {
        "NUM_FEATURES_ORIGINAL": num_features_original,
        "WINDOW_SIZE": window_size,
    }
    model_params = {
        "DIM_EMBEDDING": 256,
        "DIM_FEEDFORWARD": 1021,
        "NUM_HEADS": 4,
        "NUM_LAYERS": 1,
        "DROPOUT": 0.2,
    }
    model = AlphaPortfolio(model_args, model_params).to(DEVICE)
    final_scores = model(x)
    print(final_scores)




from finol.utils import load_config

model_dict = {}


class ModelSelector:
    def __init__(self, load_dataset_output):
        self.config = load_config()
        self.load_dataset_output = load_dataset_output

        if "AlphaPortfolio" in self.config["MODEL_PARAMS"] and self.config["MODEL_NAME"] == "AlphaPortfolio":
            from finol.model_layer.AlphaPortfolio import AlphaPortfolio
            model_dict["AlphaPortfolio"] = AlphaPortfolio
            self.model_args = {
                "NUM_FEATURES_ORIGINAL": self.load_dataset_output["NUM_FEATURES_ORIGINAL"],
                "WINDOW_SIZE": self.load_dataset_output["WINDOW_SIZE"],
            }
            self.model_params = {
                "NUM_LAYERS": self.config["MODEL_PARAMS"]["AlphaPortfolio"]["NUM_LAYERS"],
                "DIM_EMBEDDING": self.config["MODEL_PARAMS"]["AlphaPortfolio"]["DIM_EMBEDDING"],
                "DIM_FEEDFORWARD": self.config["MODEL_PARAMS"]["AlphaPortfolio"]["DIM_FEEDFORWARD"],
                "NUM_HEADS": self.config["MODEL_PARAMS"]["AlphaPortfolio"]["NUM_HEADS"],
                "DROPOUT": self.config["MODEL_PARAMS"]["AlphaPortfolio"]["DROPOUT"],
            }

        if "CNN" in self.config["MODEL_PARAMS"] and self.config["MODEL_NAME"] == "CNN":
            from finol.model_layer.CNN import CNN
            model_dict["CNN"] = CNN
            self.model_args = {
                "NUM_FEATURES_ORIGINAL": self.load_dataset_output["NUM_FEATURES_ORIGINAL"],
                "WINDOW_SIZE": self.load_dataset_output["WINDOW_SIZE"],
            }
            self.model_params = {
                "OUT_CHANNELS": self.config["MODEL_PARAMS"]["CNN"]["OUT_CHANNELS"],
                "KERNEL_SIZE": self.config["MODEL_PARAMS"]["CNN"]["KERNEL_SIZE"],
                "STRIDE": self.config["MODEL_PARAMS"]["CNN"]["STRIDE"],
                "HIDDEN_SIZE": self.config["MODEL_PARAMS"]["CNN"]["HIDDEN_SIZE"],
                "DROPOUT": self.config["MODEL_PARAMS"]["CNN"]["DROPOUT"],
            }

        if "DNN" in self.config["MODEL_PARAMS"] and self.config["MODEL_NAME"] == "DNN":
            from finol.model_layer.DNN import DNN
            model_dict["DNN"] = DNN
            self.model_args = {
                "NUM_FEATURES_AUGMENTED": self.load_dataset_output["NUM_FEATURES_AUGMENTED"],
            }
            self.model_params = {
                "NUM_LAYERS": self.config["MODEL_PARAMS"]["DNN"]["NUM_LAYERS"],
                "HIDDEN_SIZE": self.config["MODEL_PARAMS"]["DNN"]["HIDDEN_SIZE"],
                "DROPOUT": self.config["MODEL_PARAMS"]["DNN"]["DROPOUT"],
            }

        if "LSRE-CAAN" in self.config["MODEL_PARAMS"] and self.config["MODEL_NAME"].startswith("LSRE-CAAN"):
            from finol.model_layer.LSRE_CAAN import LSRE_CAAN
            model_dict["LSRE-CAAN"] = LSRE_CAAN
            model_dict["LSRE-CAAN-d"] = LSRE_CAAN
            model_dict["LSRE-CAAN-dd"] = LSRE_CAAN
            self.model_args = {
                "NUM_FEATURES_ORIGINAL": self.load_dataset_output["NUM_FEATURES_ORIGINAL"],
                "WINDOW_SIZE": self.load_dataset_output["WINDOW_SIZE"],
            }
            self.model_params = {
                "NUM_LAYERS": self.config["MODEL_PARAMS"]["LSRE-CAAN"]["NUM_LAYERS"],
                "NUM_LATENTS": self.config["MODEL_PARAMS"]["LSRE-CAAN"]["NUM_LATENTS"],
                "LATENT_DIM": self.config["MODEL_PARAMS"]["LSRE-CAAN"]["LATENT_DIM"],
                "CROSS_HEADS": self.config["MODEL_PARAMS"]["LSRE-CAAN"]["CROSS_HEADS"],
                "LATENT_HEADS": self.config["MODEL_PARAMS"]["LSRE-CAAN"]["LATENT_HEADS"],
                "CROSS_DIM_HEAD": self.config["MODEL_PARAMS"]["LSRE-CAAN"]["CROSS_DIM_HEAD"],
                "LATENT_DIM_HEAD": self.config["MODEL_PARAMS"]["LSRE-CAAN"]["LATENT_DIM_HEAD"],
                "DROPOUT": self.config["MODEL_PARAMS"]["LSRE-CAAN"]["DROPOUT"],
            }

        if "LSTM" in self.config["MODEL_PARAMS"] and self.config["MODEL_NAME"] == "LSTM":
            from finol.model_layer.LSTM import LSTM
            model_dict["LSTM"] = LSTM
            self.model_args = {
                "NUM_FEATURES_ORIGINAL": self.load_dataset_output["NUM_FEATURES_ORIGINAL"],
                "WINDOW_SIZE": self.load_dataset_output["WINDOW_SIZE"],
            }
            self.model_params = {
                "NUM_LAYERS": self.config["MODEL_PARAMS"]["LSTM"]["NUM_LAYERS"],
                "HIDDEN_SIZE": self.config["MODEL_PARAMS"]["LSTM"]["HIDDEN_SIZE"],
                "DROPOUT": self.config["MODEL_PARAMS"]["LSTM"]["DROPOUT"],
            }

        if "RNN" in self.config["MODEL_PARAMS"] and self.config["MODEL_NAME"] == "RNN":
            from finol.model_layer.RNN import RNN
            model_dict["RNN"] = RNN
            self.model_args = {
                "NUM_FEATURES_ORIGINAL": self.load_dataset_output["NUM_FEATURES_ORIGINAL"],
                "WINDOW_SIZE": self.load_dataset_output["WINDOW_SIZE"],
            }
            self.model_params = {
                "NUM_LAYERS": self.config["MODEL_PARAMS"]["RNN"]["NUM_LAYERS"],
                "HIDDEN_SIZE": self.config["MODEL_PARAMS"]["RNN"]["HIDDEN_SIZE"],
                "DROPOUT": self.config["MODEL_PARAMS"]["RNN"]["DROPOUT"],
            }

        if "Transformer" in self.config["MODEL_PARAMS"] and self.config["MODEL_NAME"] == "Transformer":
            from finol.model_layer.Transformer import Transformer
            model_dict["Transformer"] = Transformer
            self.model_args = {
                "NUM_FEATURES_ORIGINAL": self.load_dataset_output["NUM_FEATURES_ORIGINAL"],
                "WINDOW_SIZE": self.load_dataset_output["WINDOW_SIZE"],
            }
            self.model_params = {
                "NUM_LAYERS": self.config["MODEL_PARAMS"]["Transformer"]["NUM_LAYERS"],
                "DIM_EMBEDDING": self.config["MODEL_PARAMS"]["Transformer"]["DIM_EMBEDDING"],
                "DIM_FEEDFORWARD": self.config["MODEL_PARAMS"]["Transformer"]["DIM_FEEDFORWARD"],
                "NUM_HEADS": self.config["MODEL_PARAMS"]["Transformer"]["NUM_HEADS"],
                "DROPOUT": self.config["MODEL_PARAMS"]["Transformer"]["DROPOUT"],
            }

    def select_model(self, sampled_params=None):
        model_cls = model_dict.get(self.config["MODEL_NAME"], None)
        if model_cls is None:
            raise ValueError(f"Invalid model: {self.config['MODEL_NAME']}. Supported models are: {model_dict}")

        if self.config["TUNE_PARAMETERS"] and sampled_params != None:
            self.model_params = sampled_params

        model = model_cls(self.model_args, self.model_params).to(self.config["DEVICE"])
        return model



import torch

from finol.utils import load_config


class CriterionSelector:
    def __init__(self):
        self.config = load_config()

    def LOG_WEALTH(self, portfolios, labels):
        daily_returns = torch.sum(portfolios * labels, dim=-1)
        log_returns = torch.log(torch.clamp(daily_returns, min=1e-6))
        loss = - torch.mean(log_returns)
        return loss

    def LOG_WEALTH_L2_DIVERSIFICATION(self, portfolios, labels):
        daily_returns = torch.sum(portfolios * labels, dim=-1)
        log_returns = torch.log(torch.clamp(daily_returns, min=1e-6))

        # Calculate the similarity loss
        # We use the L2 norm to measure the difference between consecutive predictions
        diff = portfolios[1:, :] - portfolios[:-1, :]
        diff_loss = torch.norm(diff, p=2, dim=-1)

        # By minimizing the L2 norm of the difference vectors, you will tend to get two vectors that are similar, since the
        # L2 norm measures the Euclidean distance between vectors. When the L2 norm is small, it means that the difference
        # between two vectors is small.
        loss = - torch.mean(log_returns) + self.config["LAMBDA_L2"] * torch.mean(diff_loss)
        return loss

    def L2_DIVERSIFICATION(self, portfolios, labels):
        # Calculate the similarity loss
        # We use the L2 norm to measure the difference between consecutive predictions
        diff = portfolios[1:, :] - portfolios[:-1, :]
        diff_loss = torch.norm(diff, p=2, dim=-1)

        # By minimizing the L2 norm of the difference vectors, you will tend to get two vectors that are similar, since the
        # L2 norm measures the Euclidean distance between vectors. When the L2 norm is small, it means that the difference
        # between two vectors is small.
        loss = torch.mean(diff_loss)
        return loss

    def LOG_WEALTH_L2_CONCENTRATION(self, portfolios, labels):
        daily_returns = torch.sum(portfolios * labels, dim=-1)
        log_returns = torch.log(torch.clamp(daily_returns, min=1e-6))

        # Calculate the similarity loss
        # We use the L2 norm to measure the difference between consecutive predictions
        diff = portfolios[1:, :] - portfolios[:-1, :]
        diff_loss = torch.norm(diff, p=2, dim=-1)

        # By minimizing the L2 norm of the difference vectors, you will tend to get two vectors that are similar, since the
        # L2 norm measures the Euclidean distance between vectors. When the L2 norm is small, it means that the difference
        # between two vectors is small.
        loss = - torch.mean(log_returns) - self.config["LAMBDA_L2"] * torch.mean(diff_loss)
        return loss

    def L2_CONCENTRATION(self, portfolios, labels):
        # Calculate the similarity loss
        # We use the L2 norm to measure the difference between consecutive predictions
        diff = portfolios[1:, :] - portfolios[:-1, :]
        diff_loss = torch.norm(diff, p=2, dim=-1)

        # By minimizing the L2 norm of the difference vectors, you will tend to get two vectors that are similar, since the
        # L2 norm measures the Euclidean distance between vectors. When the L2 norm is small, it means that the difference
        # between two vectors is small.
        loss = - torch.mean(diff_loss)
        return loss

    def SHARPE_RATIO(self, portfolios, labels):
        daily_returns = torch.sum(portfolios * labels, dim=-1)
        # log_returns = torch.log(torch.clamp(daily_returns, min=1e-6))
        excess_returns = daily_returns  # risk_free_rate = 0

        mean_excess_return = torch.mean(excess_returns)
        std_excess_return = torch.std(excess_returns)

        sharpe_ratio = mean_excess_return / (std_excess_return + 1e-6)
        loss = - sharpe_ratio
        return loss


    # def SORTINO_RATIO(self, portfolios, labels):
    #     daily_returns = torch.sum(portfolios * labels, dim=-1)
    #     excess_returns = daily_returns
    #
    #     mean_excess_return = torch.mean(excess_returns)
    #     # Downside risk
    #     downside_risk = torch.sqrt(torch.mean(torch.clamp(target_return - daily_returns, min=0) ** 2))
    #
    #     sortino_ratio = mean_excess_return / downside_risk
    #     loss = -sortino_ratio  # 最大化 Sortino Ratio 等价于最小化其负值
    #     return loss

    def VOLATILITY(self, portfolios, labels):
        daily_returns = torch.sum(portfolios * labels, dim=-1)
        volatility = torch.std(daily_returns)
        loss = volatility
        return loss

    def __call__(self, preds, labels):
        criterion = getattr(self, self.config["CRITERION_NAME"])

        return criterion(preds, labels)
